@@ -177,7 +177,6 @@ static void gnome_print_pdf_font_print_lastchar (GnomePrintPdf *pdf, GnomePrintP
 static gint gnome_print_pdf_print_bpath (GnomePrintPdf *pdf, const ArtBpath *bpath);
 
 static gchar *gnome_print_pdf_get_date (void);
-static gchar *gnome_print_pdf_id_new (GnomePrintPdf *pdf);
 
 static gint  gnome_print_pdf_graphic_mode_set (GnomePrintPdf *pdf, PdfGraphicMode mode);
 static gint  gnome_print_pdf_page_fprintf (GnomePrintPdf *pdf, const char *format, ...);
@@ -338,7 +337,7 @@ gnome_print_pdf_construct (GnomePrintContext *ctx)
 		}
 	}
 
-	gnome_print_pdf_fprintf (pdf, "%cPDF-1.3" EOL, '%');
+	gnome_print_pdf_fprintf (pdf, "%cPDF-1.4" EOL, '%');
 	gnome_print_pdf_fprintf (pdf, "%c%c%c%c%c" EOL, '%',
 				 181, 237, 174, 251); /* [1] page 22-23 */
 	
@@ -439,21 +438,93 @@ gnome_print_pdf_stroke (GnomePrintContext *ctx, const ArtBpath *bpath)
 	return GNOME_PRINT_OK;
 }
 
+/* todo: this and pdf_image() can probably be merged into a single function */
+static void
+gnome_print_pdf_rgba_image_mask (GnomePrintContext *pc, gint object_num, const guchar *data,
+				 gint width, gint height, gint rowstride)
+{
+	GnomePrintPdfPage *page;
+	GnomePrintPdf *pdf;
+	gint hex_size, row, ret = 0;
+	guchar *hex;
+	gint length_object_num, len = 0;
+
+	pdf = GNOME_PRINT_PDF (pc);
+
+	/* The image object */
+	gnome_print_pdf_object_start (pdf, object_num, FALSE);
+	gnome_print_pdf_fprintf (pdf,
+				 "/Type /XObject" EOL
+				 "/Subtype /Image" EOL
+				 "/Name /Im%d" EOL
+				 "/Width %d" EOL
+				 "/Height %d" EOL
+				 "/ColorSpace /DeviceGray" EOL
+				 "/BitsPerComponent 8" EOL,
+				 object_num,
+				 width,
+				 height);
+	length_object_num = gnome_print_pdf_object_new (pdf);
+	gnome_print_pdf_fprintf (pdf, "/Length %d 0 R" EOL, length_object_num);
+
+	/* FIXME: We are not doing stream compresion/encoding yet. */
+	gnome_print_pdf_fprintf (pdf, "/Filter [/ASCIIHexDecode ]" EOL);
+	gnome_print_pdf_fprintf (pdf, ">>" EOL);
+	gnome_print_pdf_fprintf (pdf, "stream" EOL);
+
+	/* The image data */
+	hex = g_new (guchar, gnome_print_encode_hex_wcs (width * 3));
+	for (row = 0; row < height; row++) {
+		gint x;
+		hex_size = 0;
+		
+		for (x = 0; x < rowstride; x += 4) {
+			hex_size += gnome_print_encode_hex (data + (row * rowstride) + x + 3,
+							    hex + hex_size, 1);
+			hex_size--; /* encode_hex always (annoyingly) appends a \n at the end */
+		}
+
+		len += gnome_print_pdf_print_sized (pdf, hex, hex_size * sizeof (gchar));
+		len += gnome_print_pdf_fprintf (pdf, EOL);
+	}
+	g_free (hex);
+	/* END: The image data */
+	
+	gnome_print_pdf_fprintf (pdf,  "endstream" EOL "endobj" EOL);
+	ret += gnome_print_pdf_object_end (pdf, object_num, TRUE);
+
+	/* The length object,
+	 * which we write at the end so that we don't
+	 * have to pre-calculate the size of the stream. This allows us to write the
+	 * pdf in one pass and not have to keep the image in memory.
+	 */
+	gnome_print_pdf_object_start (pdf, length_object_num, TRUE);
+	gnome_print_pdf_fprintf (pdf,
+				 "%d 0 obj" EOL
+				 "%d" EOL
+				 "endobj" EOL,
+				 length_object_num, len);
+	gnome_print_pdf_object_end (pdf, length_object_num, TRUE);	
+
+	/* Reference on the page stream to the image that has just been created */
+	page = pdf->pages->data;
+
+	/* And now add it to the images used for this page */
+	page->images = g_list_prepend (page->images, GINT_TO_POINTER (object_num));
+}
+
 /* FIXME: Change the other gnome-print image functions to match this prototype */
 static int
 gnome_print_pdf_image (GnomePrintContext *pc, const gdouble *ctm, const guchar *data,
 		       gint width, gint height, gint rowstride, gint bytes_per_pixel)
-{	GnomePrintPdfPage *page;
+{
+	GnomePrintPdfPage *page;
 	GnomePrintPdf *pdf;
-	gint hex_size, row, object_num, ret = 0;
+	gint hex_size, row, object_num, ret = 0, mask_object_num;
 	guchar *hex;
 	gint length_object_num, len = 0;
 
-	if (bytes_per_pixel == 4) {
-		g_print ("Printing images with alpha has not beein implemented for the PDF backend\n");
-		return GNOME_PRINT_ERROR_UNKNOWN;
-	}
-	g_return_val_if_fail (1 == bytes_per_pixel || 3 == bytes_per_pixel, GNOME_PRINT_ERROR_UNKNOWN);
+	g_return_val_if_fail (1 == bytes_per_pixel || 3 == bytes_per_pixel || 4 == bytes_per_pixel, GNOME_PRINT_ERROR_UNKNOWN);
 
 	pdf = GNOME_PRINT_PDF (pc);
 
@@ -461,6 +532,7 @@ gnome_print_pdf_image (GnomePrintContext *pc, const gdouble *ctm, const guchar *
 
 	/* The image object */
 	object_num = gnome_print_pdf_object_new (pdf);
+
 	gnome_print_pdf_object_start (pdf, object_num, FALSE);
 	gnome_print_pdf_fprintf (pdf,
 				 "/Type /XObject" EOL
@@ -476,20 +548,35 @@ gnome_print_pdf_image (GnomePrintContext *pc, const gdouble *ctm, const guchar *
 				 bytes_per_pixel == 1 ? "DeviceGray" : "DeviceRGB");
 	length_object_num = gnome_print_pdf_object_new (pdf);
 	gnome_print_pdf_fprintf (pdf, "/Length %d 0 R" EOL, length_object_num);
+
+	if (bytes_per_pixel == 4) {		
+		mask_object_num = gnome_print_pdf_object_new (pdf);
+		gnome_print_pdf_fprintf (pdf, "/SMask %d 0 R" EOL, mask_object_num);
+	}
+
 	/* FIXME: We are not doing stream compresion/encoding yet. */
-#if 0	
-	gnome_print_pdf_write_compression_filters (pc, compr_type);
-#else
 	gnome_print_pdf_fprintf (pdf, "/Filter [/ASCIIHexDecode ]" EOL);
-#endif	
 	gnome_print_pdf_fprintf (pdf, ">>" EOL);
 	gnome_print_pdf_fprintf (pdf, "stream" EOL);
 
 	/* The image data */
 	hex = g_new (guchar, gnome_print_encode_hex_wcs (width * bytes_per_pixel));
 	for (row = 0; row < height; row++) {
-		hex_size = gnome_print_encode_hex (data + row * rowstride,
-						   hex, width * bytes_per_pixel);
+		hex_size = 0;
+
+		if (bytes_per_pixel != 4) {
+			hex_size = gnome_print_encode_hex (data + row * rowstride,
+							   hex, width * bytes_per_pixel);
+		} else {
+			gint x;
+
+			for (x = 0; x < rowstride; x += 4) {
+				hex_size += gnome_print_encode_hex (data + (row * rowstride) + x,
+								    hex + hex_size, 3);				
+				hex_size--; /* encode_hex always (annoyingly) appends a \n at the end */
+			}
+		}
+
 		len += gnome_print_pdf_print_sized (pdf, hex, hex_size * sizeof (gchar));
 		len += gnome_print_pdf_fprintf (pdf, EOL);
 	}
@@ -518,6 +605,7 @@ gnome_print_pdf_image (GnomePrintContext *pc, const gdouble *ctm, const guchar *
 		page->used_grayscale_images = TRUE;
 	else
 		page->used_color_images = TRUE;
+
 	gnome_print_pdf_page_fprintf (pdf, "q" EOL);
 	gnome_print_pdf_page_print_double (pdf, "%g",
 				      ctm[0]);
@@ -537,12 +625,14 @@ gnome_print_pdf_image (GnomePrintContext *pc, const gdouble *ctm, const guchar *
 	gnome_print_pdf_page_print_double (pdf, "%g",
 				      ctm[5]);
 	gnome_print_pdf_page_fprintf (pdf, " cm" EOL);
-	gnome_print_pdf_page_fprintf (pdf, "0 0 m" EOL);
 	gnome_print_pdf_page_fprintf (pdf, "/Im%d Do" EOL, object_num);
 	gnome_print_pdf_page_fprintf (pdf, "Q" EOL);
 
 	/* And now add it to the images used for this page */
 	page->images = g_list_prepend (page->images, GINT_TO_POINTER (object_num));
+
+	if (bytes_per_pixel == 4)
+		gnome_print_pdf_rgba_image_mask (pc, mask_object_num, data, width, height, rowstride);
 	
 	return ret;
 }
@@ -1122,7 +1212,7 @@ gnome_print_pdf_close_write_last_objects (GnomePrintPdf *pdf)
 	gint object_num_info;
 	gint object_num_catalog;
 	gint xref_offset;
-	gchar *date, *producer, *id;
+	gchar *date, *producer;
 	gint ret = 0, objects;
 
 	/* Write the default Halftone object */
@@ -1198,14 +1288,12 @@ gnome_print_pdf_close_write_last_objects (GnomePrintPdf *pdf)
 	}
 	
 	/* Write trailer */
-	id = gnome_print_pdf_id_new (pdf);
 	gnome_print_pdf_fprintf (pdf,
 				 "trailer" EOL
 				 "<<" EOL
 				 "/Size %d" EOL
 				 "/Root %d 0 R" EOL
 				 "/Info %d 0 R" EOL
-				 "/ID [<%s><%s>]" EOL
 				 ">>" EOL
 				 "startxref" EOL
 				 "%d" EOL
@@ -1213,9 +1301,8 @@ gnome_print_pdf_close_write_last_objects (GnomePrintPdf *pdf)
 				 objects,
 				 object_num_catalog,
 				 object_num_info,
-				 id, id, xref_offset,
+				 xref_offset,
 				 '%', '%');
-	g_free (id);
 	
 	return ret;
 }
@@ -1626,9 +1713,37 @@ gnome_print_pdf_font_print_descriptor (GnomePrintPdf *pdf, GnomePrintPdfFont *fo
 	return GNOME_PRINT_OK;
 }
 		
+/* 
+ * See
+ * http://partners.adobe.com/public/developer/en/pdf/PDFReference15_v5.pdf
+ * Section Section 5.5.3 Font Subsets.
+ * BaseFont name for subfont needs to be 'XXXXXX+Original Font's BaseFont name'
+ */
+
+static guchar *
+gnome_print_pdf_get_subfont_name (guchar *basefont, guint unique_subfont_id) {
+	int i;
+	guchar *full_basefont;
+
+	if (!basefont)
+		return NULL;
+	/* 6 letters, a '+' and ending NUL */
+	full_basefont = (guchar *)g_malloc (strlen(basefont) + 6 + 1 + 1);
+
+	for (i = 0; i < 6; i++) {
+		int offset = unique_subfont_id%26;
+		unique_subfont_id /= 26;
+		full_basefont[i] = 'A'+offset;
+	}
+	full_basefont[6] = '+';
+	strcpy(full_basefont + 7, basefont);
+	return (full_basefont);
+}
+
 void
 gnome_print_embed_pdf_font (GnomePrintPdf *pdf, GnomePrintPdfFont *font) {
 	gint object_number_descriptor = 0;
+	guchar *basefont_name;
 	gboolean plan_b = FALSE; /* TRUE if we could not embed the font */
 
 	if (!font->is_basic_14) {
@@ -1643,6 +1758,7 @@ gnome_print_embed_pdf_font (GnomePrintPdf *pdf, GnomePrintPdfFont *font) {
 		}
 	}
 
+	basefont_name = gnome_print_pdf_get_subfont_name ((guchar *)gnome_font_face_get_ps_name (font->face), font->object_number);
 	/* Write the object data */
 	gnome_print_pdf_object_start (pdf, font->object_number, FALSE);
 	gnome_print_pdf_fprintf  (pdf,
@@ -1651,8 +1767,9 @@ gnome_print_embed_pdf_font (GnomePrintPdf *pdf, GnomePrintPdfFont *font) {
 				  "/BaseFont /%s" EOL
 				  "/Name /F%i" EOL,
 				  font->is_type_1 ? "Type1" : "TrueType",
-				  plan_b ? (const guchar *) "Times-Roman" : gnome_font_face_get_ps_name (font->face),
+				  plan_b ? (const guchar *) "Times-Roman" : basefont_name,
 				  font->object_number);
+	g_free (basefont_name);
 
 	if (!font->is_basic_14) {
 		gnome_print_pdf_fprintf  (pdf,
@@ -1893,26 +2010,6 @@ gnome_print_pdf_get_date (void)
 	return date;
 }
 
-static gchar*
-gnome_print_pdf_id_new (GnomePrintPdf *pdf)
-{
-	gchar *id;
-	
-	id = g_strdup_printf ("%.12d%.2d%.12d%.6ld",
-			      (gint) time(NULL),
-			      95,
-			      abs (pdf->offset),
-			      (long) getpid());
-	if (strlen (id) != 32) {
-		g_warning ("Error while creating pdf_id. [%s]\n", id);
-		if (id != NULL)
-			g_free (id);
-		id = g_strdup ("00ff00ff00ff00ff00ff00ff00ff00ff");
-	}
-
-	return id;
-}
-
 
 gint
 gnome_print_pdf_fprintf (GnomePrintPdf *pdf, const char *format, ...)
@@ -2012,14 +2109,11 @@ static gint
 gnome_print_pdf_page_print_double (GnomePrintPdf *pdf, 
 				   const gchar *format, gdouble x)
 {
- 	gchar *text;
+ 	gchar text[G_ASCII_DTOSTR_BUF_SIZE];
 	gint len;
 
-	text = g_new (gchar, G_ASCII_DTOSTR_BUF_SIZE);
 	g_ascii_formatd (text, G_ASCII_DTOSTR_BUF_SIZE, format, x);
-
 	len = gnome_print_pdf_page_write (pdf, text);
-	g_free (text);
 
 	return len;
 }
